/*****************************************************************************
 *  FTools - Freenet Client Protocol Tools
 *
 *  Copyright (C) 2002 Juergen Buchmueller <juergen@pullmoll.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *****************************************************************************/
#include "ft.h"
#include "ftsite.h"
#include "ftfile.h"
#include "ftmime.h"
#include "ftfcp.h"
#include "ftlog.h"

char *progname = NULL;
char *progpath = NULL;
char *fcpchk = NULL;
char *fcpdel = NULL;
char *fcpput = NULL;
char *fcpget = NULL;

int verbose = L_NORMAL;
char fcphost[256] = "127.0.0.1";
int fcpport = 8481;
int htl = 5;
int wiggle = 0;
int revision = 1;
int get_force = 0;
//size_t chunksize = 0x100000;	/* 1MB chunks */
size_t chunksize = 262144;	/* 256KB chunks */
char mimetype[256] = "auto";
int raw = 0;
int sf_threads = 3;
int in_threads = 5;
int attempts = 25;
char default_file[256] = "index.html";
int date_offs = 0;
int date_incr = 86400;
int date_future = 1;
int meta_only = 0;
int delete = 0;

void version(void)
{
	printf ("FT (Freenet Tools) version " VERSION "\n");
}

void usage(void)
{
	if (0 == strcasecmp(progname, "fthello") ||
		0 == strcasecmp(progname, "fthello.exe")) {
		fprintf(stderr, "usage: %s [options]\n", progname);
	} else if (0 == strcasecmp(progname, "ftkey") ||
		0 == strcasecmp(progname, "ftkey.exe")) {
		fprintf(stderr, "usage: %s [options]\n", progname);
	} else if (0 == strcasecmp(progname, "ftchk") ||
		0 == strcasecmp(progname, "ftchk.exe")) {
		fprintf(stderr, "usage: %s [options] file\n", progname);
	} else if (0 == strcasecmp(progname, "ftdel") ||
		0 == strcasecmp(progname, "ftdel.exe")) {
		fprintf(stderr, "usage: %s [options] CHK@\n", progname);
	} else if (0 == strcasecmp(progname, "ftget") ||
		0 == strcasecmp(progname, "ftget.exe")) {
		fprintf(stderr, "usage: %s [options] URI [file]\n", progname);
	} else if (0 == strcasecmp(progname, "ftput") ||
		0 == strcasecmp(progname, "ftput.exe")) {
		fprintf(stderr, "usage: %s [options] URI file\n", progname);
	} else if (0 == strcasecmp(progname, "ftgetsite") ||
		0 == strcasecmp(progname, "ftgetsite.exe")) {
		fprintf(stderr, "usage: %s [options] SSK@<key>/<site> dir\n",
			progname);
		fprintf(stderr, "<key>\t\tpublic key of a freesite\n");
		fprintf(stderr, "<site>\t\tbase name of this site\n");
	} else if (0 == strcasecmp(progname, "ftputsite") ||
		0 == strcasecmp(progname, "ftputsite.exe")) {
		fprintf(stderr, "usage: %s [options] SSK@<key>/<site> dir\n",
			progname);
		fprintf(stderr, "<key>\t\tprivate key obtained by e.g. 'ftkey'\n");
		fprintf(stderr, "<site>\t\tbase name of this site\n");
	} else {
		fprintf(stderr, "usage: %s [options] <command> [URI] [file|dir]\n",
			progname);
		fprintf(stderr, "<command> can be one of:\n");
		fprintf(stderr, "hello\t\t\t\tcontact node and print NodeHello info\n");
		fprintf(stderr, "key\t\t\t\tgenerate a private/public key pair\n");
		fprintf(stderr, "chk file\t\t\tcalculate CHK@ for file\n");
		fprintf(stderr, "put URI file\t\t\tstore file under the given URI\n");
		fprintf(stderr, "get URI [file]\t\t\tretrieve URI [and store to file]\n");
		fprintf(stderr, "getsite SSK@<key>/<site> dir\tretrieve directory tree from site\n");
		fprintf(stderr, "putsite SSK@<key>/<site> dir\tinsert directory tree as site\n");
	}
	fprintf(stderr, "parameter description:\n");
	fprintf(stderr, "URI\t\t\tCHK@, [KSK@]somename or SSK@<key> (key = private key\n\t\t\t\tfor inserts; key = public key for requests)\n");
	fprintf(stderr, "file\t\t\t[/some/path/]filename[.ext]\n");
	fprintf(stderr, "options can be none, one or several of:\n");
	fprintf(stderr, "-V|--version\t\tprint the FT version number\n");
	fprintf(stderr, "-v|--verbose <n>\t0=none 1=error 2=normal 3=minor 4=debug (default %d)\n",
		verbose);
	fprintf(stderr, "-n|--address <addr>\tset FCP server address (default %s)\n",
		fcphost);
	fprintf(stderr, "-p|--port <port>\tset FCP port (default %d)\n",
		fcpport);
	fprintf(stderr, "-s|--size <bytes>\tset split file size (default 0x%x)\n",
		chunksize);
	fprintf(stderr, "-l|--htl <n>\t\tset Hops To Live (default %d)\n",
		htl);
	fprintf(stderr, "-r|--revision <n>\tset revision (default %d)\n",
		revision);
	fprintf(stderr, "-i|--insert-threads <n>\tnumber of insert threads (default %d)\n",
		in_threads);
	fprintf(stderr, "-a|--attempts <n>\tnumber of insertion attempts for each file (default %d)\n",
		attempts);
	fprintf(stderr, "-t|--threads <n>\tsplit file threads (default %d)\n",
		sf_threads);
	if (0 == strcasecmp(progname, "ft") ||
		0 == strcasecmp(progname, "ft.exe") ||
		0 == strcasecmp(progname, "ftputsite") ||
		0 == strcasecmp(progname, "ftputsite.exe")) {
		fprintf(stderr, "-D|--default <n>\tdefault document name (%s)\n",
			default_file);
		fprintf(stderr, "-d|--no-dbr\t\tno date based redirect (DBR); insert edition based site\n");
		fprintf(stderr, "-f|--future <n>\t\tinsert n days (time units) in the future (default %d)\n", date_future);
		fprintf(stderr, "--meta-only\tdo not insert data but metadata only: GenerateCHK (default %s)\n", meta_only ? "on" : "off");
		fprintf(stderr, "--delete\tdelete keys from the local store before insertion (default %s)\n", delete ? "on" : "off");
	}
	if (0 == strcasecmp(progname, "ft") ||
		0 == strcasecmp(progname, "ft.exe") ||
		0 == strcasecmp(progname, "ftget") ||
		0 == strcasecmp(progname, "ftget.exe")) {
		fprintf(stderr, "-R|--raw\t\traw mode: don't follow redirect(s)\n");
	}
}

int results(fcp_t *f, int rc, const char *uri)
{
	if (NULL != f) {
		if (0 == rc) {
			if (f->seen_key_collision) {
				printf(KEYCOLLISION ": %s\n", f->uri ? f->uri : uri);
			} else {
				printf(SUCCESS ": %s\n", f->uri ? f->uri : uri);
			}
		} else {
			if (f->seen_data_not_found) {
				printf(DATANOTFOUND ": %s\n", f->uri ? f->uri : uri);
			} else if (f->seen_route_not_found) {
				printf(ROUTENOTFOUND ": %s\n", f->uri ? f->uri : uri);
			} else if (f->seen_uri_error) {
				printf(URIERROR ": %s\n", f->uri ? f->uri : uri);
			} else if (f->seen_size_error) {
				printf(SIZEERROR ": %s\n", f->uri ? f->uri : uri);
			} else {
				printf(GENERALERROR ": %s\n", f->uri ? f->uri : uri);
			}
		}
	} else {
		printf(UNKNOWNERROR ": %s\n", uri);
	}
	return -1;
}

int main(int argc, char **argv)
{
	fcp_t *f;
	char *cmd = NULL;
	char *uri = NULL;
	char *filename = NULL;
	char *infoname = NULL;
	char *prg, *ext;
	struct stat st;
	int rc;
	int i;

	/* Construct path to ourself so that we may reinvoke ourself.  But
	   if we were invoked simply as 'ft' (relying on $PATH) then there
	   should be no path supplied at all for recursive invocations; rely
	   on $PATH once again.  Thus, we want a trailing slash after the
	   progpath if (and only if) there is a non-empty progpath.  Under
	   no circumstances do we want to use $PWD here. -GW */
	progpath = strdup(argv[0]);
	prg = strrchr(progpath, '/');
	if (NULL != prg) {
		*prg++ = '/';
		*prg++ = '\0';
		progname = prg;
	} else {
		progname = strdup(argv[0]);
		free(progpath);
		progpath = "";
	}
	ext = strrchr(progname, '.');
	if (NULL == ext) {
		ext = "";
	}

	fcpchk = calloc(strlen(progpath) + 16, 1);
	sprintf(fcpchk, "%sftchk%s", progpath, ext);
	if (0 != stat(fcpchk, &st)) {
		sprintf(fcpchk, "%sft%s chk", progpath, ext);
	}

	fcpdel = calloc(strlen(progpath) + 16, 1);
	sprintf(fcpdel, "%sftdel%s", progpath, ext);
	if (0 != stat(fcpdel, &st)) {
		sprintf(fcpdel, "%sft%s del", progpath, ext);
	}

	fcpget = calloc(strlen(progpath) + 16, 1);
	sprintf(fcpget, "%sftget%s", progpath, ext);
	if (0 != stat(fcpchk, &st)) {
		sprintf(fcpchk, "%sft%s get", progpath, ext);
	}

	fcpput = calloc(strlen(progpath) + 16, 1);
	sprintf(fcpput, "%sftput%s", progpath, ext);
	if (0 != stat(fcpchk, &st)) {
		sprintf(fcpchk, "%sft%s put", progpath, ext);
	}

	if (0 == strcasecmp(progname, "fthello") ||
		0 == strcasecmp(progname, "fthello.exe")) {
		cmd = "hello";
	} else if (0 == strcasecmp(progname, "ftchk") ||
		0 == strcasecmp(progname, "ftchk.exe")) {
		cmd = "chk";
	} else if (0 == strcasecmp(progname, "ftdel") ||
		0 == strcasecmp(progname, "ftdel.exe")) {
		cmd = "del";
	} else if (0 == strcasecmp(progname, "ftkey") ||
		0 == strcasecmp(progname, "ftkey.exe")) {
		cmd = "key";
	} else if (0 == strcasecmp(progname, "ftget") ||
		0 == strcasecmp(progname, "ftget.exe")) {
		cmd = "get";
	} else if (0 == strcasecmp(progname, "ftput") ||
		0 == strcasecmp(progname, "ftput.exe")) {
		cmd = "put";
	} else if (0 == strcasecmp(progname, "ftgetsite") ||
		0 == strcasecmp(progname, "ftgetsite.exe")) {
		cmd = "getsite";
	} else if (0 == strcasecmp(progname, "ftputsite") ||
		0 == strcasecmp(progname, "ftputsite.exe")) {
		cmd = "putsite";
	}

	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-' && argv[i][1] != '\0') {
#undef	OPT
#define	OPT	"-n"
			if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					strcpy(fcphost, &argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					strcpy(fcphost, argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"--address"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					strcpy(fcphost, &argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					strcpy(fcphost, argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"-p"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					fcpport = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					fcpport = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"--port"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					fcpport = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					fcpport = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"-v"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					verbose = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					verbose = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"--verbose"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					verbose = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					verbose = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"-V"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				version();
				return 0;
#undef	OPT
#define	OPT	"--version"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				version();
				return 0;
#undef	OPT
#define	OPT	"-i"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					in_threads = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					in_threads = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"--insert-threads"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					in_threads = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					in_threads = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"-t"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					sf_threads = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					sf_threads = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"--threads"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					sf_threads = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					sf_threads = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"-s"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					chunksize = strtoul(&argv[i][sizeof(OPT) - 1], NULL, 0);
				} else if (i + 1 < argc) {
					i++;
					chunksize = strtoul(argv[i], NULL, 0);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"--size"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					chunksize = strtoul(&argv[i][sizeof(OPT) - 1], NULL, 0);
				} else if (i + 1 < argc) {
					i++;
					chunksize = strtoul(argv[i], NULL, 0);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"-l"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					htl = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					htl = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"--htl"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					htl = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					htl = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"-w"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					wiggle = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					wiggle = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"--wiggle"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					wiggle = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					wiggle = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"-r"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					revision = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					revision = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"--revision"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					revision = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					revision = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"-d"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				date_incr = 0;
#undef	OPT
#define	OPT	"--no-dbr"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				date_incr = 0;
#undef	OPT
#define	OPT	"-f"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					date_future = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					date_future = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"--future"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					date_future = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					date_future = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"-D"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					strncpy(default_file, &argv[i][sizeof(OPT) - 1],
						sizeof(default_file));
				} else if (i + 1 < argc) {
					i++;
					strncpy(default_file, argv[i], sizeof(default_file));
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"--default"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					strncpy(default_file, &argv[i][sizeof(OPT) - 1],
						sizeof(default_file));
				} else if (i + 1 < argc) {
					i++;
					strncpy(default_file, argv[i], sizeof(default_file));
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"-a"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1))
{
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					attempts = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					attempts = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"--attempts"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1))
{
				if (strlen(argv[i]) > sizeof(OPT) - 1) {
					attempts = atoi(&argv[i][sizeof(OPT) - 1]);
				} else if (i + 1 < argc) {
					i++;
					attempts = atoi(argv[i]);
				} else {
					fprintf(stderr, "missing argument for %s\n", OPT);
					usage();
					return 1;
				}
#undef	OPT
#define	OPT	"-R"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				raw = 1;
#undef	OPT
#define	OPT	"--raw"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				raw = 1;
#undef	OPT
#define	OPT	"--meta-only"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				meta_only = 1;
#undef	OPT
#define	OPT	"--delete"
			} else if (0 == strncmp(argv[i], OPT, sizeof(OPT) - 1)) {
				delete = 1;
#undef	OPT
			} else {
				fprintf(stderr, "unknown switch '%s'\n",
					argv[i]);
				usage();
				return 1;
			}
		} else if (NULL == cmd) {
			cmd = argv[i];
		} else if (NULL == uri) {
			uri = argv[i];
		} else if (NULL == filename) {
			filename = argv[i];
		} else if (NULL == infoname) {
			infoname = argv[i];
		} else {
			usage();
			return 1;
		}
	}

	/* No command: display usage info */
	if (NULL == cmd) {
		usage();
		return 1;
	}

	srand(time(NULL));

	/* No URI and not hello or key commands: display usage info */
	if (NULL == uri && strcmp(cmd, "hello") && strcmp(cmd, "key")) {
		usage();
		return 1;
	}
	if (NULL == filename) {
		filename = "/dev/null";
	}
	/* No informational name given: use the uri */
	if (NULL == infoname) {
		infoname = uri;
	}

	if (0 != fcp_new(&f, fcphost, fcpport)) {
		printf(GENERALERROR ": connecting to %s:%d\n", fcphost, fcpport);
		return 1;
	}

	if (0 == strcasecmp(cmd, "hello")) {
		rc = fcp_hello(f);

		if (0 == rc) {
			printf("Node Protocol : %s\n", f->seen_protocol);
			printf("Node Version  : %s\n", f->seen_node);
		}

	} else if (0 == strcasecmp(cmd, "get")) {
		rc = fcp_get(f, uri, htl);

		/* successful and not raw mode? */
		if (0 == rc && 0 == raw) {

			/* if we have no data, the metadata should tell us what to do */
			if (0 == raw && NULL == f->data) {
				if (0 != (rc = get_file(&f, uri, htl, wiggle, filename, infoname))) {
					return results(f, rc, uri);
				}
			}

			/* now we should have data (or it was already output) */
			if (NULL != f->data) {
				FILE *fp;

				printf(SUCCESS ": %s\n", f->uri ? f->uri : uri);
				if (0 == strcmp(filename, "-")) {
					fp = stdout;
				} else {
					fp = fopen(filename, "wb");
				}
				if (NULL == fp) {
					fprintf(stderr, "cannot create '%s' (%s)\n",
						filename, strerror(errno));
					return 1;
				}
				if (f->datasize != fwrite(f->data, 1, f->datasize, fp)) {
					fprintf(stderr, "cannot write %d bytes to '%s' (%s)\n",
						f->datasize, filename, strerror(errno));
					return 1;
				}
				if (fp != stdout) {
					fclose(fp);
				}
				fp = NULL;

			} else {
				rc = results(f, rc, uri);
			}

		} else if (0 == rc && 0 != raw) {

			/* we should have metadata */
			if (NULL != f->meta) {
				FILE *fp;

				printf(SUCCESS ": %s\n", f->uri ? f->uri : uri);
				if (0 == strcmp(filename, "-")) {
					fp = stdout;
				} else {
					fp = fopen(filename, "wb");
				}
				if (NULL == fp) {
					fprintf(stderr, "cannot create '%s' (%s)\n",
						filename, strerror(errno));
					return 1;
				}
				if (f->metasize != fwrite(f->meta, 1, f->metasize, fp)) {
					fprintf(stderr, "cannot write %d bytes to '%s' (%s)\n",
						f->metasize, filename, strerror(errno));
					return 1;
				}
				if (fp != stdout) {
					fclose(fp);
				}
				fp = NULL;

			} else {
				printf(GENERALERROR ": %s\n", f->uri ? f->uri : uri);
			}

		} else {
			rc = results(f, rc, uri);
		}

	} else if (0 == strcasecmp(cmd, "put")) {
		rc = put_file(&f, uri, htl, attempts, wiggle, filename,
				(strcasecmp(mimetype, "auto")) ?
					mimetype : mimetype_guess(filename), 1);

		rc = results(f, rc, uri);

	} else if (0 == strcasecmp(cmd, "chk")) {
		/* we don't need no URI for generate CHK */
		if (0 == strcmp(filename, "/dev/null") && NULL != uri) {
			filename = uri;
			uri = "CHK@";
		}

		rc = chk_file(&f, filename,
			(strcasecmp(mimetype, "auto")) ?
					mimetype : mimetype_guess(filename), 1);

		rc = results(f, rc, uri);

	} else if (0 == strcasecmp(cmd, "del")) {
		rc = fcp_del(f, uri);
		rc = results(f, rc, uri);

	} else if (0 == strcasecmp(cmd, "key")) {
		rc = fcp_key(f);
		if (0 == rc) {
			printf("Public Key  : %s\n", f->public_key);
			printf("Private Key : %s\n", f->private_key);
		}

	} else if (0 == strcasecmp(cmd, "getsite")) {
		rc = fcp_get(f, uri, htl);

		rc = get_site(&f, uri, htl, wiggle, filename, revision, get_force);
		rc = results(f, rc, uri);

	} else if (0 == strcasecmp(cmd, "putsite")) {
		rc = put_site(&f, uri, htl, attempts, wiggle, filename, revision);
		rc = results(f, rc, uri);
	}

	fcp_free(&f);

	return rc;
}
